/*****************************************************************
play & record audio

$Id: plax.c,v 1.1.1.1 1998/12/29 21:26:36 purnhage Exp $

Source file: plax.c

Authors:
HP    Heiko Purnhagen, Uni Hannover <purnhage@tnt.uni-hannover.de>

Changes:
16-sep-98   HP      first version
18-sep-98   HP      mono/stereo and sample rate conversion
21-sep-98   HP      fixed numSample%2=0 for stereo rate conversion
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include <libtsp.h>		/* AFsp audio file library */
#include <libtsp/AFpar.h>	/* AFsp audio file library - definitions */

#include "common.h"		/* common module */
#include "cmdline.h"		/* command line module */
#include "austream.h"		/* audio i/o streams */
#include "playrec.h"		/* audio play & record */
#include "rateconv.h"		/* audio sample rate conversion */


/* ---------- declarations ---------- */

#define PROGTITLE "plax V0.1 21-sep-98"

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define STRLEN (255)

#define FS_MAX (48000.0)
#define BUF_TIME (1.)		/* length of buffer in seconds */

#define RAWENV "RAWAUDIOFILE"
#define RAWDEFAULT "integer16,0,48000.,native,1,1.0"


/* ---------- variables ---------- */

/* command line module */

static char *progName = "<no program name>";
static char *outFileName;
static int outFileNameUsed;
static int outNumChan;
static int outNumChanUsed;
static float outFSample;
static int outFSampleUsed;
static int recNumChan;
static int recNumChanUsed;
static float recFSample;
static int recFSampleUsed;
static int convert;
static int htaps1;
static float a;
static float fc;
static float fd;
static float gain;
static float duration;
static float buftime;
static char *rawFormat;
static int rawFormatUsed;
static int debugLevel;
static int *varArgIdx;

static CmdLinePara paraList[] = {
  {&varArgIdx,NULL,"<input files (- = stdin) (dflt: audio device)>"},
  {NULL,NULL,NULL}
};

static CmdLineSwitch switchList[] = {
  {"h",NULL,NULL,NULL,NULL,"print help"},
  {"o",&outFileName,"%s",NULL,&outFileNameUsed,
   "output file name (- = stdout) (dflt: audio device)"},
  {"c",&outNumChan,"%d",NULL,&outNumChanUsed,
   "output: number of channels (dflt: as input or 1)"},
  {"s",&outFSample,"%f",NULL,&outFSampleUsed,
   "output: sampling frequency [Hz] (dflt: as input or 8000.0)"},
  {"rc",&recNumChan,"%d",NULL,&recNumChanUsed,
   "rec: number of channels (0=current) (dflt: as output)"},
  {"rs",&recFSample,"%f",NULL,&recFSampleUsed,
   "rec: sampling frequency [Hz] (0=current) (dflt: as output)"},
  {"x",&convert,NULL,NULL,NULL,
   "enable channel & sample rate conversion"},
  {"ht",&htaps1,"%d","-1",NULL,
   "num taps"},
  {"al",&a,"%f","-1.0",NULL,
   "alpha for Kaiser window"},
  {"fc",&fc,"%f","-1.0",NULL,
   "6 dB cutoff freq / input bandwidth"},
  {"fd",&fd,"%f","-1.0",NULL,
   "100 dB cutoff freq / input bandwidth"},
  {"g",&gain,"%f","1.0",NULL,
   "amplitude gain"},
  {"t",&duration,"%f","0.0",NULL,
   "duration [sec] (0=unlimited)"},
  {"b",&buftime,"%f","0.2",NULL,
   "buffer time"},
  {"f",&rawFormat,"%s",NULL,&rawFormatUsed,
   "raw input file format\n"
   "(dflt: $" RAWENV " / \"" RAWDEFAULT "\")"},
  {"d",&debugLevel,"%d","0",NULL,
   "debug level (0=off)"},
  {NULL,NULL,NULL,NULL,NULL,NULL}
};

static int stop = 0;


/* ---------- local functions ---------- */

void stopproc (int sig)
{
  stop = 1;
}


/* ---------- main ---------- */

int main (int argc, char *argv[])
{
  int listIdx;
  char inFileName[STRLEN];
  char fileName[STRLEN];
  AFILE *inaf,*outaf;
  AuStream *inas,*outas;
  PRStream *inpr,*outpr;
  RCBuf *rc;
  int result;
  int RCdebugLevel;

  long maxBufLen,bufLen,convLen,bufTrans;
  long inBufUsed,outBufUsed,tmpBufUsed;
  long numClip,totClip;
  short *bufS;
  float *bufF;
  short *outBufS;
  float *outBufF;
  long i,tmp;

  int inNumChan;
  long inNumChanL;
  float inFSample;
  long fileNumSample;
  long inNumSamp;
  long outNumSamp;
  int devNumChan;
  float devFSample;
  long maxNumSamp;
  int fmt;

  typedef enum {AUDEV, AUSTREAM, AUFILE} autype;
  autype inty,outty;

  /* evaluate command line  */
  CmdLineInit(0);
  result = CmdLineEval(argc,argv,paraList,switchList,1,&progName);
  if (result) {
    if (result==1) {
      fprintf(stderr,"%s: %s\n",progName,PROGTITLE);
      CmdLineHelp(progName,paraList,switchList,stderr);
      exit (1);
    }
    else
      CommonExit(1,"command line error (\"-h\" for help)");
  }
  AuInit((debugLevel>>2)&3);
  PRInit((debugLevel>>4)&3);
  RCdebugLevel = (debugLevel>>6)&3;
  debugLevel &= 3;

  if (debugLevel)
    fprintf(stderr,"%s: %s\n",progName,PROGTITLE);
  if (debugLevel>1)
    fprintf(stderr,"debugLevel=%d\n",debugLevel);

  if (rawFormatUsed)
    AFsetNH(rawFormat);
  else {
    if (getenv(RAWENV))
      AFsetNH("$" RAWENV);
    else
      AFsetNH(RAWDEFAULT);
  }

  maxBufLen = (long)(BUF_TIME*FS_MAX*2);
  if(!(bufF = (float *)malloc(maxBufLen*sizeof(float))))
    CommonExit(1,"failed to allocate bufF");
  if(!(bufS = (short *)malloc(maxBufLen*sizeof(short))))
    CommonExit(1,"failed to allocate bufS");
  outBufS = bufS;

  if (outFileNameUsed && varArgIdx[0]>=0 && varArgIdx[1]>=0)
    CommonExit(1,"max one input file allowed when using -o");
  
  signal(SIGINT,stopproc);

  /* process all files on command line or record from audio device */
  listIdx = 0;
  while (!stop && (outFileNameUsed || varArgIdx[listIdx] >= 0)) {
    /* open input */
    if (varArgIdx[listIdx]>=0) {
      if (ComposeFileName(argv[varArgIdx[listIdx]],
			  0,NULL,NULL,inFileName,STRLEN))
	CommonExit(1,"composed input file name too long \"%s\"",argv[listIdx]);
      if (strcmp(inFileName,"-")==0) {
	inty = AUSTREAM;
	if (!(inas=AuOpenRead(inFileName,&inNumChan,&inFSample,
			      &fileNumSample)))
	  CommonExit(1,"can not open input \"%s\"",inFileName);
      }
      else {
	inty = AUFILE;
	if (!(inaf=AFopenRead(inFileName,&fileNumSample,&inNumChanL,
			      &inFSample,
			      (debugLevel>1)?stderr:(FILE*)NULL)))
	  CommonExit(1,"can not open input \"%s\"",inFileName);
	inNumChan = inNumChanL;
      }
    }
    else {
      inty = AUDEV;
      fileNumSample = -1;
      inNumChan = (outNumChanUsed)?outNumChan:1;
      inFSample = (outFSampleUsed)?outFSample:8000.0;
      if (recNumChanUsed)
	inNumChan = recNumChan;
      if (recFSampleUsed)
	inFSample = recFSample;
      if (!(inpr=PROpenRead(0,&inNumChan,&inFSample)))
	CommonExit(1,"can not open input audio device");
      if (inNumChan<=0 && outNumChanUsed)
	inNumChan = outNumChan;
      if (inFSample<=0 && outFSampleUsed)
	inFSample = outFSample;
      if (inNumChan<=0 || inFSample<=0)
	CommonExit(1,"input audio device configuration troubles");
    }
    if (debugLevel>1)
      fprintf(stderr,
	      "input=%s inNumChan=%d inFSample=%.1f fileNumSample=%ld\n",
	      (inty==AUDEV)?"audiodevice":inFileName,
	      inNumChan,inFSample,fileNumSample);
    
    /* open output */
    if (!outNumChanUsed)
      outNumChan = inNumChan;
    if (!outFSampleUsed)
      outFSample = inFSample;

    if (outFileNameUsed) {
      if (ComposeFileName(outFileName,0, NULL,NULL,fileName,STRLEN))
	CommonExit(1,"composed output file name too long \"%s\"",outFileName);
      if (strcmp(fileName,"-")==0) {
	outty = AUSTREAM;
	if (!(outas=AuOpenWrite(fileName,outNumChan,outFSample)))
	  CommonExit(1,"can not open output \"%s\"",fileName);
      }
      else {
	outty = AUFILE;
	fmt = FD_INT16;
	if (strstr(fileName,"au")!=NULL ||
	    strstr(fileName,"snd")!=NULL)
	  fmt += FW_SUN;
	else if (strstr(fileName,"wav")!=NULL)
	  fmt += FW_WAVE;
	else if (strstr(fileName,"aif")!=NULL)
	  fmt += FW_AIFF_C;
	else if (strstr(fileName,"raw")!=NULL)
	  fmt +=  FW_NH_EB;	/* big-endian */
	else
	  CommonExit(1,"unkown output audio in format");
	if (!(outaf=AFopenWrite(fileName,fmt,outNumChan,outFSample,
				(debugLevel>1)?stderr:(FILE*)NULL)))
	  CommonExit(1,"can not open output \"%s\"",fileName);
      }
    }
    else {
      outty = AUDEV;
      devNumChan = outNumChan;
      devFSample = outFSample;
      if (!(outpr=PROpenWrite(0,&devNumChan,&devFSample)))
	CommonExit(1,"can not open output audio device");
      if (devNumChan != outNumChan || devFSample != outFSample)
	CommonWarning("can not configure output audio device as requested");
    }
    if (debugLevel>1)
      fprintf(stderr,"output=%s outNumChan=%d outFSample=%.1f\n",
	      (outty==AUDEV)?"audiodevice":fileName,outNumChan,outFSample);

    /* copy samples */
    if (convert && (inNumChan > 2 || outNumChan > 2))
      CommonExit(-1,"file conversion only for mono/stereo");
    bufLen = (long)(inNumChan*inFSample*buftime+.5);
    bufLen = (bufLen/inNumChan)*inNumChan;
    rc = NULL;
    if (convert && (inFSample!=outFSample || fc>=0 || fd>=0)) {
      if (!(rc=RateConvInit(RCdebugLevel,outFSample/(float)inFSample,
			    min(inNumChan,outNumChan),htaps1,a,fc,fd,
			    &convLen)))
	CommonExit(1,"can not init sample rate conversion");
      bufLen = convLen;
      if (inNumChan==2 && outNumChan==1)
	bufLen *= 2;
    }
    bufLen = min(bufLen,maxBufLen);
    if (debugLevel>2)
      fprintf(stderr,"bufLen=%ld maxBufLen=%ld\n",bufLen,maxBufLen);

    totClip = 0;
    inNumSamp = 0;
    outNumSamp = 0;
    maxNumSamp = (long)(inNumChan*inFSample*duration+.5);
    do {
      bufTrans = (maxNumSamp)?maxNumSamp-inNumSamp:bufLen;
      if (bufTrans>bufLen)
	bufTrans = bufLen;

      switch (inty) {
      case AUSTREAM:
	inBufUsed = AuReadData(inas,bufS,bufTrans);
	for (i=0; i<inBufUsed; i++)
	  bufF[i] = bufS[i];
	break;
      case AUFILE:
	inBufUsed = AFreadData(inaf,inNumSamp,bufF,bufTrans);
	break;
      case AUDEV:
	inBufUsed = PRReadData(inpr,bufS,bufTrans);
	for (i=0; i<inBufUsed; i++)
	  bufF[i] = bufS[i];
	break;
      }
      inNumSamp += inBufUsed;

      outBufUsed = inBufUsed;
      outBufF = bufF;
      if (convert) {
	if (inNumChan==2 && outNumChan==1) {
	  outBufUsed = inBufUsed/2;
	  for (i=0; i<outBufUsed; i++)
	    bufF[i] = (bufF[2*i]+bufF[2*i+1])/2;
	}
	if (rc) {
	  while (outBufUsed<convLen)
	    bufF[outBufUsed++] = 0;
	  outBufUsed = RateConv(rc,bufF,outBufUsed,&outBufF);
	  tmpBufUsed = (long)(inNumSamp/inNumChan/inFSample*
			      outFSample*outNumChan)-outNumSamp;
	  if (outNumChan==2) {
	    if (inNumChan==1)
	      tmpBufUsed /= 2;
	    else
	      tmpBufUsed = (tmpBufUsed/outNumChan)*outNumChan;
	  }
	  if (tmpBufUsed<outBufUsed)
	    outBufUsed = tmpBufUsed;
	}
	if (inNumChan==1 && outNumChan==2) {
	  outBufUsed *= 2;
	  for (i=outBufUsed-1; i>=0; i--)
	    bufF[i] = outBufF[i/2];
	  outBufF = bufF;
	}
      }
      if (outBufUsed>maxBufLen)
	CommonExit(-1,"output buffer size troubles");
      numClip = 0;
      for (i=0; i<outBufUsed; i++) {
	tmp = ((long)(gain*outBufF[i]+32768.5))-32768;
	if (tmp>32767) {
	  tmp = 32767;
	  numClip++;
	}
	if (tmp<-32768) {
	    tmp = -32768;
	    numClip++;
	}
	outBufS[i] = tmp;
      }
      totClip += numClip;

      if (debugLevel) {
	if (debugLevel<=2) {
	  fprintf(stderr,".");
	  fflush(stderr);
	}
	else
	  fprintf(stderr,"inBufUsed=%ld outBufUsed=%ld numClip=%ld\n",
		  inBufUsed,outBufUsed,numClip);
      }

      switch (outty) {
      case AUSTREAM:
	AuWriteData(outas,outBufS,outBufUsed);
	break;
      case AUFILE:
	AFwriteData(outaf,outBufF,outBufUsed);
	break;
      case AUDEV:
	PRWriteData(outpr,outBufS,outBufUsed);
	break;
      }
      outNumSamp += outBufUsed;

    } while (!stop && inBufUsed);
      
    if (rc)
      RateConvFree(rc);
      
    /* close input */
    switch (inty) {
    case AUSTREAM:
      AuClose(inas);
      break;
    case AUFILE:
      AFclose(inaf);
      break;
    case AUDEV:
      PRClose(inpr);
      break;
    }

    /* close output */
    switch (outty) {
    case AUSTREAM:
      AuClose(outas);
      break;
    case AUFILE:
      AFclose(outaf);
      break;
    case AUDEV:
      PRClose(outpr);
      break;
    }

    if (debugLevel<=2) {
      fprintf(stderr," \n");
      fflush(stderr);
    }
    if (debugLevel)
      fprintf(stderr,
	      "in: %ld samples (%.3f sec)   "
	      "out: %ld samples (%.3f sec)\n",
	      inNumSamp,inNumSamp/(inNumChan*inFSample),
	      outNumSamp,outNumSamp/(outNumChan*outFSample));
    if (totClip)
      CommonWarning("%ld samples clipped!",totClip);

    if (outFileNameUsed)
      outFileNameUsed = 0;
    if (varArgIdx[listIdx]>=0)
      listIdx++;
  }

  CmdLineEvalFree(paraList);

  if (debugLevel>1)
    fprintf(stderr,"%s: %s\n",progName,stop?"aborted":"finished");

  exit (0);
}

/* end of plax.c */

